{% extends "sphinx_book_theme/layout.html" %} {% block fonts %}
<!-- add `style` or `link` tags with your CSS `@font-face` declarations here -->
<!-- ... and a `style` tag with setting `font-family` in `body` and `.header-style` -->
<!-- ... and optionally preload the `woff2` for snappier page loads -->
<!-- or add a `style` tag with a font fallback chain with good cross-platform coverage -->
<style>
    body {
        font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto,
            Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;
    }

    .header-style {
        font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto,
            Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;
    }
</style>
{% endblock %}
