METRIC_CV_CELL = "cv_cell"
METRIC_CV_GENE = "cv_gene"
METRIC_DIFF_EXP = "diff_exp"
METRIC_CALIBRATION = "calibration"
METRIC_ZERO_FRACTION = "zero_fraction"

DEFAULT_DE_N_TOP_GENES = 2
DEFAULT_DE_P_VAL_THRESHOLD = 0.001
DEFAULT_DE_N_TOP_GENES_OVERLAP = 100

DATA_VAR_RAW = "Raw"

UNS_NAME_RGG_RAW = "rank_genes_groups_raw"
UNS_NAME_RGG_PPC = "rank_genes_groups_ppc"
